DRWATERGITROOT <- "drwater.rcees.ac.cn/git/"
DRWATERPKGS <- c(
  "dwfun", "langeR", "rmdify", "dateR", "dfeR",
  "limvis", "uniteR", "cctda", "cctdb",
  "mapeR", "meteR", "drwateR", "uniteR"
)
# Define core packages
CORE <- c("dwfun", "langeR", "rmdify", "dateR", "dfeR", "limvis", "uniteR")

#' Download package from drwateR Git repository
#'
#' Downloads a package from the drwateR Git repository.
#'
#' @param pkgname Character; Name of the package to download.
#' @param gitroot Character; The root URL of the Git repository
#' '(default: "drwater.rcees.ac.cn/git/").
#' @param scheme Character; The URL scheme to use (default: "https").
#' @param gitrooturl Character; Full Git root URL
#' (default: constructed using scheme and gitroot).
#' @param cred Object; Credentials for Git authentication
#' '(default: `credentials::git_credential_ask`).
#'
#' @return Character; The local path of the downloaded package or NULL
#' if the download fails.
#'
#' @importFrom git2r clone
#' @importFrom credentials git_credential_ask
#'
#' @export
#'
#' @examples
#' download_drwateR("rmdify")
download_drwateR <- function(
    pkgname,
    gitroot = DRWATERGITROOT,
    scheme = "https",
    gitrooturl = paste0(scheme, "://", gitroot),
    cred = credentials::git_credential_ask(gitrooturl)) {
  pkgurl <- if (is.null(cred)) {
    paste0(gitrooturl, "drwateR/", pkgname, ".git")
  } else {
    paste0(
      "https://", cred$username, ":", cred$password, "@",
      gitroot, "drwateR/", pkgname, ".git"
    )
  }

  localpath <- file.path(tempdir(), paste0(pkgname, ".git"))

  if (dir.exists(paths = localpath)) {
    unlink(localpath, recursive = TRUE, force = FALSE, expand = TRUE)
  }

  path <- tryCatch(
    gsub("\\/.git$", "", git2r::clone(pkgurl, local_path = localpath)$path),
    error = function(e) {
      return(NULL)
    }
  )

  return(path)
}


#' Install package from drwateR Git repository
#'
#' Installs a package from the drwateR Git repository, including handling
#' dependencies.
#'
#' @importFrom utils install.packages
#' @param pkgname Character; Name of the package to install.
#' @param dependency_reinstall Logical; If TRUE, reinstall dependencies.
#' @param pkgpath Character; Local path to the downloaded package (default:
#' path returned by `download_drwateR`).
#' @param drwaterpkgs Character; A vector of package names belonging to `drwateR`.
#' @param force Logical; If TRUE, force the installation even if the package is up-to-date.
#'
#' @return Logical; TRUE if the package was successfully installed, FALSE otherwise.
#' @export
#'
#' @examples
#' install_drwateR("rmdify")
install_drwateR <- function(pkgname,
                            dependency_reinstall = FALSE,
                            pkgpath = download_drwateR(pkgname),
                            drwaterpkgs = DRWATERPKGS,
                            force = FALSE) {
  # Check if multiple packages are provided
  if (length(pkgname) > 1) {
    sapply(pkgname, function(i) {
      install_drwateR(i,
        dependency_reinstall = dependency_reinstall,
        force = force
      )
    })
    return(invisible(TRUE))
  }

  # Validate package name
  if (!(pkgname %in% drwaterpkgs)) {
    warning(pkgname, " does not belong to `drwateR`!")
    return(invisible(FALSE))
  }

  # Ensure necessary packages are installed
  if (!requireNamespace("credentials", quietly = TRUE)) {
    install.packages("credentials")
  }
  if (!requireNamespace("git2r", quietly = TRUE)) {
    install.packages("git2r")
  }

  # Validate the package path
  if (is.null(pkgpath) || !dir.exists(pkgpath)) {
    stop(pkgname, " has not been successfully downloaded!")
  }

  # Handle dependencies
  depdf <- dependency_drwateR(pkgpath)
  if (nrow(depdf) > 0) {
    install_dependency(depdf, dependency_reinstall = dependency_reinstall)
  }

  # Check for version updates
  if (requireNamespace(pkgname, quietly = TRUE)) {
    remoteversion <- unname(read.dcf(file.path(
      pkgpath,
      "DESCRIPTION"
    ))[, "Version"])
    localversion <- gsub(
      "^‘|’$", "",
      as.character(packageVersion(pkgname))
    )

    if (remoteversion == localversion && !force) {
      message("No update in remote!")
      unlink(pkgpath, recursive = TRUE, force = FALSE, expand = TRUE)
      return(invisible(FALSE))
    }
  }

  # Install the package
  result <- tryCatch(
    system(paste("R CMD INSTALL", shQuote(pkgpath)), intern = TRUE),
    error = function(e) {
      warning("Failed to install package from ", pkgpath)
      return(NULL)
    }
  )


  if (is.null(result)) {
    warning("R CMD INSTALL exited with status ", result)
    out <- FALSE
  } else {
    invisible(unlist(sapply(result, message)))
    message("Package installed successfully from ", pkgpath)
    out <- TRUE
  }

  unlink(pkgpath, recursive = TRUE, force = TRUE)
  return(invisible(out))
}

#' Install dependencies for a package
#'
#' Installs the dependencies for a specified package from various sources
#' including CRAN, GitHub, and drwateR.
#'
#' @param depdf A data frame containing information about dependency packages.
#' Must include columns `pkgname`, `source`, and `installed`.
#' @param dependency_reinstall Logical; if TRUE, reinstall all dependencies
#' regardless of their installation status.
#'
#' @importFrom devtools install_github
#' @importFrom utils install.packages
#' @export
#'
install_dependency <- function(depdf, dependency_reinstall = FALSE) {
  # Filter installed packages based on the `dependency_reinstall` flag
  if (!dependency_reinstall) {
    depdf <- depdf[!depdf$installed, ]
  }

  # Install CRAN packages
  cranpkgs <- depdf[depdf$source == "cran", "pkgname"]
  if (length(cranpkgs) > 0) {
    message("Installing CRAN packages: ", paste(cranpkgs, collapse = ", "))
    install.packages(cranpkgs)
  }

  # Install GitHub packages
  githubpkgs <- depdf[depdf$source == "github", "pkgname"]
  if (length(githubpkgs) > 0) {
    message("Installing GitHub packages: ", paste(githubpkgs, collapse = ", "))
    invisible(sapply(githubpkgs, function(pkg) {
      devtools::install_github(pkg)
    }))
  }

  # Install drwateR packages
  drwateRpkgs <- depdf[depdf$source == "drwateR", "pkgname"]
  if (length(drwateRpkgs) > 0) {
    message("Installing drwateR packages: ", paste(drwateRpkgs, collapse = ", "))
    invisible(sapply(drwateRpkgs, install_drwateR))
  }
}



#' Find all dependency packages for a given package
#'
#' This function identifies all dependency packages listed in the
#' DESCRIPTION file of a local R package source code. It categorizes
#' the dependencies based on their source: CRAN, drwateR, or GitHub.
#'
#' @param pkgpath Character; the local path to the package source code.
#' @param desc A data frame containing the content of the DESCRIPTION file.
#'              If not provided, it will be read from the specified `pkgpath`.
#' @param drwaterpkgs Character vector; the list of drwateR packages.
#'
#' @return A data.frame containing the names of dependency packages,
#'         whether they are installed, and their source.
#'
#' @export
#'
dependency_drwateR <- function(
    pkgpath,
    desc = read.dcf(file.path(pkgpath, "DESCRIPTION")),
    drwaterpkgs = DRWATERPKGS) {
  # Extract dependencies from DESCRIPTION
  depends <- if ("Depends" %in% colnames(desc)) {
    gsub(" (.*)$", "", unlist(strsplit(unname(desc[, "Depends"]), split = ",\\n")))
  } else {
    character(0)
  }

  imports <- if ("Imports" %in% colnames(desc)) {
    gsub(" (.*)$", "", unlist(strsplit(unname(desc[, "Imports"]), split = ",\\n")))
  } else {
    character(0)
  }

  remotes <- if ("Remotes" %in% colnames(desc)) {
    gsub(" (.*)$", "", unlist(strsplit(unname(desc[, "Remotes"]), split = ",\\n")))
  } else {
    character(0)
  }

  # Combine unique package names
  pkgname <- unique(c(depends, imports, remotes))
  pkgname <- pkgname[pkgname != "R"]

  # Create data frame of dependencies
  depdf <- data.frame(
    pkgname = pkgname,
    installed = sapply(pkgname, function(x) requireNamespace(x, quietly = TRUE)),
    stringsAsFactors = FALSE
  )

  # Assign sources
  depdf$source <- "cran"
  depdf$source[depdf$pkgname %in% drwaterpkgs] <- "drwateR"
  depdf$source[grepl("\\/", depdf$pkgname)] <- "github"

  return(depdf)
}

#' Update drwateR packages
#'
#' Updates one or more packages from the drwateR Git repository.
#'
#' @param pkgname Character vector; the names of packages to update.
#'                If `NULL`, updates all drwateR packages by default.
#'
#' @return Logical; `TRUE` if the packages were successfully updated,
#'         `FALSE` otherwise.
#' @export
#'
#' @examples
#' update_drwateR()
#'
update_drwateR <- function(pkgname = DRWATERPKGS) {
  # Install each package and check for success
  results <- sapply(pkgname, function(pkg) {
    tryCatch(
      {
        install_drwateR(pkg)
        TRUE # Return TRUE if installation succeeds
      },
      error = function(e) {
        warning(paste("Failed to update package:", pkg, "-", e$message))
        FALSE # Return FALSE if installation fails
      }
    )
  })
  # Return TRUE if all packages were successfully updated, FALSE otherwise
  return(invisible(all(results)))
}
